/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : 
 *        File    : adm8668_prc.c
 *        Abstract: 
 *        Date    : 
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/
#include <linux/config.h>
#include <asm/adm8668/adm8668.h>
#include <asm/adm8668/bsp_sup.h>
#include <linux/proc_fs.h>
#include <linux/compatmac.h>	// copy_from_user
#include <linux/slab.h>

#define CONFIG_IFX_GAN	/* for GAN projecc */

#ifdef CONFIG_ADM8668_HW_NAT
#include <asm/adm8668/hw_nat.h>
extern void HwNatEnable(int i);
extern int CheckHwNatEnable(void);
#endif

#ifdef CONFIG_IFX_GAN
/* LED state:
	Green off: 0x0, on: 0x1, slow flash: 0x2, fast flash: 0x4
	Red off: 0x0, on: 0x10, slow flash: 0x20, fast flash: 0x40
*/
int led_state = 0x1;
struct timer_list led_timer;
#endif

static struct proc_dir_entry * proc_adm8668_root = NULL;
void adm8668_proc_init(void);

struct proc_dir_entry * adm8668_proc_getproc(void)
{
	return proc_adm8668_root;
}

/* 
 * proc file system add function for adm8668
 */
int adm8668_proc_addproc(char *funcname, read_proc_t *hookfuncr, write_proc_t *hookfuncw)
{
	struct proc_dir_entry *pe;

	if (!proc_adm8668_root)
		adm8668_proc_init();

	if (hookfuncw == NULL)
	{
		pe = create_proc_read_entry(funcname, 0, proc_adm8668_root, hookfuncr, NULL);
		if (!pe)
		{
			printk("ERROR in creating read proc entry (%s)! \n", funcname);
			return -1;
		}
	}else
	{
		pe = create_proc_entry(funcname, S_IRUGO | S_IWUGO, proc_adm8668_root);
		if (pe)
		{
			pe->read_proc = hookfuncr;
			pe->write_proc = hookfuncw;
		}else
		{
			printk("ERROR in creating proc entry (%s)! \n", funcname);
			return -1;
		}
	}

	return 0;
}

#ifdef CONFIG_IFX_GAN
int proc_adm8668_imei0_entry(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	BOARD_CFG_T brdCfg; 

	memcpy(&brdCfg, (char *)PA2VA(ADM8668_BOARD_CFG_ADDR), sizeof(brdCfg));
	if (brdCfg.imeimagic == ADM8668_IMEI_MAGIC)
	{
		len += sprintf(buf+len, "%s\n", brdCfg.imei0);
	}
	return len; 
}

int proc_adm8668_imei1_entry(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	BOARD_CFG_T brdCfg; 

	memcpy(&brdCfg, (char *)PA2VA(ADM8668_BOARD_CFG_ADDR), sizeof(brdCfg));
	if (brdCfg.imeimagic == ADM8668_IMEI_MAGIC)
	{
		len += sprintf(buf+len, "%s\n", brdCfg.imei1);
	}
	return len;
}

static void led_isr(unsigned long data)
{
	int timer_plus = 0;

	del_timer(&led_timer);
	switch (led_state & 0xf)	/* for green LED */
	{
		case 0x2:
			GPIO_TOGGLE(0);
			timer_plus = HZ/2; 
			break;

		case 0x4:
			GPIO_TOGGLE(0);
			timer_plus = HZ/5; 
			break;

		case 0:
			GPIO_SET_LOW(0);
			break;
	}

	switch (led_state & 0xf0)
	{
		case 0x20:
			CRGPIO_TOGGLE(2);
			timer_plus = HZ/2; 
			break;
		case 0x40:
			CRGPIO_TOGGLE(2);
			timer_plus = HZ/5; 
			break;
		case 0:
			CRGPIO_SET_LOW(2);
			break;
	}

	if (timer_plus != 0)
	{
		led_timer.expires = jiffies + timer_plus;
		add_timer(&led_timer);
	}

	return;
}

static void gan_led_init(void)
{
	ADM8668_CONFIG_REG(CRGPIO_REG) |= GPIO2_OUTPUT_ENABLE;
	ADM8668_WLAN_REG(GPIO_REG) |= GPIO0_OUTPUT_ENABLE;
	init_timer(&led_timer);
	led_timer.data = (long)NULL;
	led_timer.function = led_isr;

	if ((led_state & 0x6) || (led_state & 0x60))
	{
		led_timer.expires = HZ/2;
		add_timer(&led_timer);
	}
}

#if 1
int proc_adm8668_led_write_entry(struct file *file, const char *buffer, unsigned long count, void *data)
{
	char buf[8];
	int num;

	num = (count < 8) ? count : 8;

	if (copy_from_user(buf, buffer, num))
	{
		printk("copy_from_user failed");
                return -EFAULT;
	}
	num = simple_strtoul(buf, NULL, 16);
	switch (num)
	{
		case 0:
			GPIO_SET_LOW(0);
			CRGPIO_SET_LOW(2);
			break;
		case 1:
			GPIO_SET_LOW(0);
			CRGPIO_SET_HI(2);
			break;
		case 2:
			GPIO_SET_HI(0);
			CRGPIO_SET_HI(2);
			break;
		default:
			break;
	}

	return count;
}

int proc_adm8668_led_read_entry(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	int led_state = 0;

	led_state = (ADM8668_CONFIG_REG(CRGPIO_REG) & 0x100) ? 1 : 0;
	led_state += (ADM8668_WLAN_REG(GPIO_REG) & 0x40) ? 2 : 0;
	len += sprintf(buf+len, "%s\n",
		(led_state&1) ? 
			((led_state&2) ? "ORANGE" : "GREEN") : "OFF");

	return len;
}

int proc_adm8668_button_read_entry(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	int read_once = ADM8668_CONFIG_REG(CRGPIO_REG);
	int button_flip = (read_once >> 20) & 0x3;
	int button_state = read_once & 0x3;

	len += sprintf(buf+len, "SES: %s %s\nRESET: %s %s\n",
		(button_state&2) ? "UP" : "DOWN",
		(button_flip&2) ? "FLIP" : "",
		(button_state&1) ? "UP" : "DOWN",
		(button_flip&1) ? "FLIP" : "");

	return len;
}
#endif

#endif
void adm8668_proc_init(void)
{
	int bogus;

	proc_adm8668_root = proc_mkdir("adm8668", 0);
	if (!proc_adm8668_root)
	{
		printk("ADM8668 proc initialization failed! \n");
		return;
	}
	printk("ADM8668 proc initialization okay! \n");
#ifdef CONFIG_IFX_GAN
	ADM8668_CONFIG_REG(CRGPIO_REG) = GPIO2_OUTPUT_ENABLE;
	ADM8668_WLAN_REG(GPIO_REG) = GPIO0_OUTPUT_ENABLE;

	/* inital read off of the flipper switcher on the button thingie */
	bogus = ADM8668_CONFIG_REG(CRGPIO_REG);

	adm8668_proc_addproc("led", proc_adm8668_led_read_entry, proc_adm8668_led_write_entry);
	adm8668_proc_addproc("button", proc_adm8668_button_read_entry, NULL);
//	adm8668_proc_addproc("imei0", proc_adm8668_imei0_entry, NULL);
//	adm8668_proc_addproc("imei1", proc_adm8668_imei1_entry, NULL);
#endif
}

#ifdef CONFIG_ADM8668_HW_NAT
static int HWNAT_readfile(char *page, char **start, off_t offset, int count,
                   int *eof, void *data)
{
	int i;
	int len;
	char *out = page;

	i = CheckHwNatEnable();
	out += sprintf(out,"%i",i);

	if (i == MODE_PPPOE)
		out += sprintf(out, "(PPPoE)\n");
	else if (i == MODE_NORMAL)
		out += sprintf(out, "(Normal connection)\n");
	else if (i == MODE_DISABLETURBONAT)
		out += sprintf(out, "(Disable Turbo NAT)\n");
	else
		out += sprintf(out, "(HWNAT disabled)\n");

        len = out - page;
        len -= offset;
        if (len < count) {
                *eof = 1;
                if (len <= 0) {
                        return 0;
                }
        } else {
                len = count;
        }


	*start = page + offset;	
	return len;
}

static int HWNAT_writefile(struct file *file, const char *buffer, unsigned long count,
                    void *data)

{
	char buf[8];
	int num;

	num = (count < 8) ? count : 8;

        if (copy_from_user(buf, buffer, num))
	{
		printk("copy_from_user failed");
                return -EFAULT;
	}
        buf[num] = '\0';

	HwNatEnable(buf[0]-'0');

	return num;
}


int proc_adm8668_hwnat_entry(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0, i, valid = 0;

    	len += sprintf(buf+len, "\nNAT entries (only first 50 entries):");

    	for (i=0; i<NAT_TABLE_ENTRY_NUM; i++)
    	{
        	WAIT_CR10_READY ; //wait ready bit=1

        	READ_NAT_TABLE_BY_INDEX(i);

        	WAIT_CR10_READY ; //wait ready bit=1

	        // print cr11~17 only valid bit = 1
        	if (ADM8668_CONFIG_REG(ADM8668_CR17) & CR17_ENTRY_VALID)
	        {
	        	if ((count - len) > 100) // keep the proc data < count
	        	{
        	    		len += sprintf(buf+len, "\nIDX %.3d:", i);
		            	len += sprintf(buf+len, " %.8x,", ADM8668_CONFIG_REG(ADM8668_CR11));
        		    	len += sprintf(buf+len, " %.8x,", ADM8668_CONFIG_REG(ADM8668_CR12));
	        	    	len += sprintf(buf+len, " %.8x,", ADM8668_CONFIG_REG(ADM8668_CR13));
        	    		len += sprintf(buf+len, " %.8x,", ADM8668_CONFIG_REG(ADM8668_CR14));
		            	len += sprintf(buf+len, " %.8x,", ADM8668_CONFIG_REG(ADM8668_CR15));
        		    	len += sprintf(buf+len, " %.8x,", ADM8668_CONFIG_REG(ADM8668_CR16));
	        		len += sprintf(buf+len, " %.8x", ADM8668_CONFIG_REG(ADM8668_CR17));
	        	}
	        	valid ++;
        	}
	}
	len += sprintf(buf+len, "\n");
	len += sprintf(buf+len, "Total entry number: %d\n", valid);
	
	return len;
}


int proc_adm8668_hwnat_tbls(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0, cr51;
	
    	len += sprintf(buf+len, "\n  NAT format table");
    	len += sprintf(buf+len, "\n        VA   CE    IP  Contents                     Mask bits");
    
    	cr51 = ADM8668_CONFIG_REG(ADM8668_CR51);

    	len += sprintf(buf+len, "\nEntry0: %d    %d    %3d  %.8x %.8x %.8x,  %.8x %.8x %.8x", 
        	    (cr51 & BIT_20) >> 20, (cr51 & BIT_24) >> 24, cr51 & 0x1f,
	            ADM8668_CONFIG_REG(ADM8668_CR27),
        	    ADM8668_CONFIG_REG(ADM8668_CR28),
	            ADM8668_CONFIG_REG(ADM8668_CR29),
        	    ADM8668_CONFIG_REG(ADM8668_CR30),
	            ADM8668_CONFIG_REG(ADM8668_CR31),
        	    ADM8668_CONFIG_REG(ADM8668_CR32));
    
	len += sprintf(buf+len, "\nEntry1: %d    %d    %3d  %.8x %.8x %.8x,  %.8x %.8x %.8x", 
            (cr51 & BIT_21) >> 21, (cr51 & BIT_25) >> 25, (cr51 & 0x3e0) >> 5,
            ADM8668_CONFIG_REG(ADM8668_CR33),
            ADM8668_CONFIG_REG(ADM8668_CR34),
            ADM8668_CONFIG_REG(ADM8668_CR35),
            ADM8668_CONFIG_REG(ADM8668_CR36),
            ADM8668_CONFIG_REG(ADM8668_CR37),
            ADM8668_CONFIG_REG(ADM8668_CR38));

    	len += sprintf(buf+len, "\nEntry2: %d    %d    %3d  %.8x %.8x %.8x,  %.8x %.8x %.8x", 
            (cr51 & BIT_22) >> 22, (cr51 & BIT_26) >> 26, (cr51 & 0x7c00) >> 10,
            ADM8668_CONFIG_REG(ADM8668_CR39),
            ADM8668_CONFIG_REG(ADM8668_CR40),
            ADM8668_CONFIG_REG(ADM8668_CR41),
            ADM8668_CONFIG_REG(ADM8668_CR42),
            ADM8668_CONFIG_REG(ADM8668_CR43),
            ADM8668_CONFIG_REG(ADM8668_CR44));

    	len += sprintf(buf+len, "\nEntry3: %d    %d    %3d  %.8x %.8x %.8x,  %.8x %.8x %.8x", 
            (cr51 & BIT_23) >> 23, (cr51 & BIT_27) >> 27, (cr51 & 0xf8000) >> 15,
            ADM8668_CONFIG_REG(ADM8668_CR45),
            ADM8668_CONFIG_REG(ADM8668_CR46),
            ADM8668_CONFIG_REG(ADM8668_CR47),
            ADM8668_CONFIG_REG(ADM8668_CR48),
            ADM8668_CONFIG_REG(ADM8668_CR49),
            ADM8668_CONFIG_REG(ADM8668_CR50));
    	
    	len += sprintf(buf+len, "\n");	
    	
	return len;
}

int proc_adm8668_hwnat_nogotbls(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0, cr61, cr60;
	
    	len += sprintf(buf+len, "\n  NAT NOGO table");
    	len += sprintf(buf+len, "\n        VA   CE    L4  UDP   Offset Contents  Mask bits");
    
    	cr61 = ADM8668_CONFIG_REG(ADM8668_CR61);
    	cr60 = ADM8668_CONFIG_REG(ADM8668_CR60);

    	len += sprintf(buf+len, "\nEntry0: %d    %d    %3d  %3d %8d %.8x  %.8x", 
        	    (cr61 & BIT_0) , (cr61 & BIT_4) ? 1:0, (cr61 & BIT_8)?1:0, (cr61 & BIT_12)?1:0,
        	    (cr60 & 0xff), 
	            ADM8668_CONFIG_REG(ADM8668_CR52),
        	    ADM8668_CONFIG_REG(ADM8668_CR53));
        len += sprintf(buf+len, "\nEntry1: %d    %d    %3d  %3d %8d %.8x  %.8x", 
        	    (cr61 & BIT_1)?1:0 , (cr61 & BIT_5) ? 1:0, (cr61 & BIT_9)?1:0, (cr61 & BIT_13)?1:0,
	            (cr60 >> 8) & 0xff, 
	            ADM8668_CONFIG_REG(ADM8668_CR54),
        	    ADM8668_CONFIG_REG(ADM8668_CR55));
        len += sprintf(buf+len, "\nEntry2: %d    %d    %3d  %3d %8d %.8x  %.8x", 
        	    (cr61 & BIT_2)?1:0 , (cr61 & BIT_6) ? 1:0, (cr61 & BIT_10)?1:0, (cr61 & BIT_14)?1:0,
	            (cr60 >> 16) & 0xff, 
	            ADM8668_CONFIG_REG(ADM8668_CR56),
        	    ADM8668_CONFIG_REG(ADM8668_CR57));
        len += sprintf(buf+len, "\nEntry3: %d    %d    %3d  %3d %8d %.8x  %.8x", 
        	    (cr61 & BIT_3)?1:0 , (cr61 & BIT_7) ? 1:0, (cr61 & BIT_11)?1:0, (cr61 & BIT_15)?1:0,
	            (cr60 >> 24) & 0xff, 
	            ADM8668_CONFIG_REG(ADM8668_CR58),
        	    ADM8668_CONFIG_REG(ADM8668_CR59));
    
    	
    	len += sprintf(buf+len, "\n");	
    	
	return len;
}

int proc_adm8668_hwnat_regs(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	
	len += sprintf(buf+len, "Configuration registers: ");
    	len += sprintf(buf+len, "\n  CR18: 0x%.8x, CR19: 0x%.8x, CR20: 0x%.8x, CR21: 0x%.8x,",
                ADM8668_CONFIG_REG(ADM8668_CR18), ADM8668_CONFIG_REG(ADM8668_CR19), 
                ADM8668_CONFIG_REG(ADM8668_CR20), ADM8668_CONFIG_REG(ADM8668_CR21) );
        len += sprintf(buf+len, "\n  CR24: 0x%.8x, CR25: 0x%.8x, CR26: 0x%.8x \n",
                ADM8668_CONFIG_REG(ADM8668_CR24), ADM8668_CONFIG_REG(ADM8668_CR25), 
                ADM8668_CONFIG_REG(ADM8668_CR26) );
        
    	len += sprintf(buf+len, "\nLAN port: ");
    	len += sprintf(buf+len, "\n  CSR36A: 0x%.8x, CSR36B: 0x%.8x, CSR36C: 0x%.8x,",
                ADM8668_LAN_REG(NETCSR36A), ADM8668_LAN_REG(NETCSR36B), 
                ADM8668_LAN_REG(NETCSR36C) );
    	len += sprintf(buf+len, "\n  CSR36D: 0x%.8x, CSR36E: 0x%.8x, CSR36F: 0x%.8x,",
                ADM8668_LAN_REG(NETCSR36D), ADM8668_LAN_REG(NETCSR36E), 
                ADM8668_LAN_REG(NETCSR36F) );
    	len += sprintf(buf+len, "\n  CSR36G: 0x%.8x, CSR36H: 0x%.8x, CSR36I: 0x%.8x.\n",
                ADM8668_LAN_REG(NETCSR36G), ADM8668_LAN_REG(NETCSR36H), 
                ADM8668_LAN_REG(NETCSR36I) );
    
    	len += sprintf(buf+len, "\nWAN port: ");
    	len += sprintf(buf+len, "\n  CSR36A: 0x%.8x, CSR36B: 0x%.8x, CSR36C: 0x%.8x,",
                ADM8668_WAN_REG(NETCSR36A), ADM8668_WAN_REG(NETCSR36B), 
                ADM8668_WAN_REG(NETCSR36C) );
    	len += sprintf(buf+len, "\n  CSR36D: 0x%.8x, CSR36E: 0x%.8x, CSR36F: 0x%.8x,",
                ADM8668_WAN_REG(NETCSR36D), ADM8668_WAN_REG(NETCSR36E), 
                ADM8668_WAN_REG(NETCSR36F) );
    	len += sprintf(buf+len, "\n  CSR36G: 0x%.8x, CSR36H: 0x%.8x, CSR36I: 0x%.8x.\n",
                ADM8668_WAN_REG(NETCSR36G), ADM8668_WAN_REG(NETCSR36H), 
                ADM8668_WAN_REG(NETCSR36I) );
        
        *eof = 1;
        return len;
}


void ADM8668_HwNat_proc_init(void)
{
	struct proc_dir_entry *de;
	
	if (!proc_adm8668_root)
		adm8668_proc_init();
			
	de = create_proc_entry("hw_nat", S_IRUGO | S_IWUGO, proc_adm8668_root);

	if (de)
	{
		de->read_proc = HWNAT_readfile; 
		de->write_proc = HWNAT_writefile; 	
    	}
    	
    	create_proc_read_entry("hw_nat_entry", 0, proc_adm8668_root, proc_adm8668_hwnat_entry, NULL);
	create_proc_read_entry("hw_nat_regs", 0, proc_adm8668_root, proc_adm8668_hwnat_regs, NULL);
	create_proc_read_entry("hw_nat_tbls", 0, proc_adm8668_root, proc_adm8668_hwnat_tbls, NULL);	
	create_proc_read_entry("hw_nat_nogo", 0, proc_adm8668_root, proc_adm8668_hwnat_nogotbls, NULL);
	
}
#endif

