/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : 
 *        File    : flash.c
 *        Abstract: 
 *        Date    : 
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/ 
//#include <linux/autoconf.h>
#include <asm/adm8668/adm8668.h>
#include <asm/adm8668/flash.h>
#if defined(CONFIG_ADM8668_NAND_CFG)
#include <linux/mtd/mtd.h>
#include <linux/slab.h>
static struct mtd_info *nandmtd = NULL; 
#endif

static FLASH_DESC *chips = 0;
extern int mx29lv_init(void);


/*----------------------------------------------------------------------
 * ROUTINE NAME - flash_getid	//jeansonh, modified from royc's code
*-----------------------------------------------------------------------
* DESCRIPTION:
*	Detect the flash driver deivce identify  
*	
*----------------------------------------------------------------------*/
unsigned short flash_getid(int bitoff)
{
	volatile register unsigned short  *block_Id  =  (unsigned short *)PA2VA(((ADM8668_SMEM0_BASE)+(bitoff)));
	volatile unsigned short id;
	int i;
	volatile unsigned short *p555 = (unsigned short *)PA2VA(ADM8668_SMEM0_BASE) + 0x555;
    	volatile unsigned short *p2aa = (unsigned short *)PA2VA(ADM8668_SMEM0_BASE) + 0x2AA;
	
	/***********************/
	/* Read id            */
	/***********************/
	
	//set to read id mode 
	*p555 = 0xaa; 
	*p2aa = 0x55; 
	*p555 = 0x90; 
	
	for(i=0; i<1000; i++) ;
	
	id = *block_Id;	   //read id
		
	//exit to normal mode 
	*p555 = 0xaa;
	*p2aa = 0x55;
	*p555 = 0xf0;
	
	for(i=0; i<1000; i++) ;
	
	return id;
}

/*----------------------------------------------------------------------
* ROUTINE NAME - flash_init
*-----------------------------------------------------------------------
* DESCRIPTION:
*	The routine is intialize the flash driver
*   
*----------------------------------------------------------------------*/
int flash_init()
{
#ifdef CONFIG_ADM8668_NOR_MXIC
    mx29lv_init();
#endif    	
       
    return 0;
}

/*----------------------------------------------------------------------
* ROUTINE NAME - flash_add
*-----------------------------------------------------------------------
* DESCRIPTION:
*	The routine is intialize the flash driver
*   
*----------------------------------------------------------------------*/
int flash_add(FLASH_DESC *dev)
{
    register FLASH_DESC *cp = chips;

    dev->next = 0;

    while (cp)
    {
        /* Already installed */
        if (cp == dev)
            return 0;
    
        if (cp->next == 0)
            break;
        else
            cp = cp->next;
    }

    if (cp == 0)
        chips = dev;
    else
        cp->next = dev;
    
    return 0;
}

/*----------------------------------------------------------------------
* ROUTINE NAME - flash_fit
*-----------------------------------------------------------------------
* DESCRIPTION:
*	The routine is to check the range of the input data
*   
* INPUT:
*   cp      -- flash control block
*	flash   -- Address of the block to read.
*   cells   -- cell number to check.
*----------------------------------------------------------------------*/
static int flash_fit (struct FLASH_DESC_S *cp, unsigned long addr, int *size)
{
    register int len;
    register unsigned long start = (unsigned long)cp->start;
    register unsigned long end = (unsigned long)cp->start + cp->flash_size;

    /* check range */
    if (addr < start || addr >= end)
    {
        return FLASH_NOT_FIT_IN;        /* out of range */
    }

    /* Get the size to write */
    len = (end - addr)/cp->addr_inc;
    if (*size > len)
    {
        *size = len;
        return FLASH_PARTIAL_DONE;      /* partial written */
    }
    else
    {
        switch (cp->byte_width)
        {
        case 16:
            *size = ((*size + 1)/2)*2;
            break;

        case 32:
            *size = ((*size + 3)/4)*4;
        
        default:
            break;
        }
    }

    return FLASH_OK;
}


/*----------------------------------------------------------------------
* ROUTINE NAME - flash_read
*-----------------------------------------------------------------------
* DESCRIPTION:
*	The routine is called by NV RAM driver to read NV data.
*   
*
* INPUT:
*	flash   -- Address of the block to read.
*   dst 	-- address of this data buffer.
*   size    -- data size to read from the flash.
*----------------------------------------------------------------------*/
int flash_read(char *flash, char *dst, int cells)
{
    int size = cells;
    register int rc = 0;
    register FLASH_DESC *cp = chips;    

    while (cp && cells > 0)
    {
        if (flash_fit(cp, (unsigned long)flash, &size) != FLASH_NOT_FIT_IN)
        {
            rc = (*cp->read)(cp, flash, dst, size);
            if (rc == FLASH_ERROR)
                break;

            flash += size * cp->addr_inc;
            dst   += size;
            cells -= size;
        }

        size = cells;
        cp = cp->next;
    }

    return rc;
}


/*----------------------------------------------------------------------
* ROUTINE NAME - flash_write
*-----------------------------------------------------------------------
* DESCRIPTION:
*	The routine is called by NV RAM driver or the image updating module 
*   to update the date content.
*
* INPUT:
*	flash   -- Address of the block to wirte.
*   src	    -- Source address of this image.
*   size    -- image size to write to the flash.
*----------------------------------------------------------------------*/
int flash_write(char *flash, char *src, int cells)
{
    int size = cells;
    register int rc = 0;
    register FLASH_DESC *cp = chips;    

    /* if size > 4MB, addr should be started @ 0xb0000000 */
    if (cp->flash_size > 0x00400000)
    	flash -= 0x0fc00000;
    
    while (cp && cells > 0)
    {
        if (flash_fit(cp, (unsigned long)flash, &size) != FLASH_NOT_FIT_IN)
        {
            rc = (*cp->write)(cp, flash, src, size);
            if (rc == FLASH_ERROR)
                break;

            flash += size * cp->addr_inc;
            src   += size;
            cells -= size;
        }

        size = cells;
        cp = cp->next;
    }

    return rc;
}


/*----------------------------------------------------------------------
* ROUTINE NAME - flash_erase
*-----------------------------------------------------------------------
* DESCRIPTION:
*	The routine is called by NV RAM driver or the image updating module 
*   to erase the blocks, whereafter this block can accept new data
*   written.
*
* INPUT:
*	flash   -- Address of the flash start address to be erased
*   size    -- size of the flash portion to be erased
*----------------------------------------------------------------------*/
int flash_erase(char *flash, int cells)
{
    int size = cells;
    register int rc = 0;
    register FLASH_DESC *cp = chips;    

    /* if size > 4MB, addr should be started @ 0xb0000000 */
    if (cp->flash_size > 0x00400000)
    	flash -= 0x0fc00000;
    
    while (cp && cells > 0)
    {
        if (flash_fit(cp, (unsigned long)flash, &size) != FLASH_NOT_FIT_IN)
        {
            rc = (*cp->erase)(cp, flash, size);
            if (rc == FLASH_ERROR)
                break;

            flash += size * cp->addr_inc;
            cells -= size;
        }
		size = cells;
        cp = cp->next;
    }

    return rc;
}

#ifdef CONFIG_ADM8668_NAND_CFG
/* for nand flash */
int adm8668_nand_init(void)
{
  struct mtd_info *mtd;
  
  mtd = get_mtd_device(NULL, 0);
  if (mtd)
  {
      printk("MTD device found!\n");
      nandmtd = mtd;
      
      return 0;
  }
  
  return 1;
}

int adm8668_nand_read(char *dst, char *src, int len)
{
  int rtval;
  
  if (!nandmtd)
  {
    rtval = adm8668_nand_init();
    if (rtval)
    { 
    	printk("ERROR finding NAND mtd device! \n");
    	return 1;
    }
  }
  
  rtval = nandmtd->read(nandmtd, (unsigned int)src, len, &rtval, dst);
  if (rtval)
  	printk("NAND read failed! (%d)\n", rtval); 
  	
  return rtval;
}

int adm8668_nand_write(char *dst, char *src, int len)
{
  int rtval;
  
  len = 0x400;
  if (!nandmtd)
  {
    rtval = adm8668_nand_init();
    if (rtval)
    {
      printk("ERROR finding NAND mtd device! \n");
      return 1;
    }
  }
  
  rtval = nandmtd->write(nandmtd, (unsigned int)dst, len, &rtval, src);  
  
  if (rtval)
  	printk("NAND read failed! (%d)\n", rtval); 
  	
  return rtval;
}

int adm8668_nand_erase(char *dst, int len)
{
  struct erase_info *einfo;
  int rtval;
  
  if (!nandmtd)
  {
    rtval = adm8668_nand_init();
    if (rtval)
    {
      printk("ERROR finding NAND mtd device! \n");
      return 1;
    }
  }
  
  einfo = kmalloc(sizeof(struct erase_info), GFP_KERNEL);
  memset(einfo, 0, sizeof(struct erase_info));
  
  einfo->mtd = nandmtd;
  einfo->addr = (unsigned int)dst;
  einfo->len = 16384;
  
  rtval = nandmtd->erase(nandmtd, einfo);
  if (rtval)
    printk("MTD erase ERROR! (%d)\n", rtval);   
  else
    printk("MTD erase from %x len %x okay! \n", einfo->addr, einfo->len);
    
  kfree(einfo);
  
  return rtval;
}
#endif	// CONFIG_ADM8668_NAND_CFG
