/******************************************************************************
**
** FILE NAME    : gan_wdt.c
** PROJECT      : GAN
** MODULES      : 
**
** DATE         : 5 NOV 2007
** AUTHOR       : Jeffrey Huang
** DESCRIPTION  : GAN WATCHDOG Driver
** COPYRIGHT    :       Copyright (c) 2006
**                      Infineon Technologies AG
**                      Am Campeon 1-12, 85579 Neubiberg, Germany
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date        $Author         $Comment
** 5 NOV 2007  Jeffrey.H        Initiate Version
*******************************************************************************/


#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include <asm/adm8668/adm8668.h>

#if CONFIG_MODVERSIONS==1
# include <linux/modversions.h>
#endif


#if 1
#define DEBUG(args...) do {;} while(0)
#else
#define DEBUG   printk
#endif


#define ADM8668_CR7	0x1C 	
#define POLLING_PERIOD 4*HZ
#define GAN_WDT_VERSION "0.0.1"
#define BIT_31 (1<<31)

struct timer_list gan_wdt_timer;

static void gan_wdt(unsigned long data)           
{
    unsigned int cr7;
    //clear CR7 	
    cr7 =  ADM8668_CONFIG_REG(ADM8668_CR7);
    DEBUG("gan_wdt is invoked and CR7 is %x\n",cr7); 
    mod_timer(&gan_wdt_timer,jiffies + POLLING_PERIOD);
}



int __init gan_wdt_init(void)
{
    int ret;


    printk("The GAN WDT version is "GAN_WDT_VERSION " The Built TIME is %s\n",__TIME__);
    // Watchdog Reset Enable
    ADM8668_CONFIG_REG(ADM8668_CR7) = BIT_31;
    init_timer(&gan_wdt_timer);
    gan_wdt_timer.data = (long)NULL;
    gan_wdt_timer.function = gan_wdt;
    gan_wdt_timer.expires =  jiffies + POLLING_PERIOD;
    add_timer(&gan_wdt_timer);
  
    return 0;
}


