/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : Daniel Liu
 *        File    : irq.c
 *        Abstract: 
 *        Date    : 2003.3.4
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/ 
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/pm.h>
#include <linux/irq.h>
#include <asm/mipsregs.h>
#include <asm/gdb-stub.h>
#include <asm/adm8668/adm8668.h>


extern void breakpoint(void);

extern int setup_irq(unsigned int irq, struct irqaction *irqaction);

extern irq_desc_t irq_desc[];
extern asmlinkage void mipsIRQ(void);

extern int mips_int_lock(void);
extern int mips_int_unlock(int s);

void enable_adm8668_irq(unsigned int irq);
void disable_adm8668_irq(unsigned int irq);

/*
 * irq enable
 */
static __inline void _irq_enable(int irql)
{
	ADM8668_INTC_REG(IRQ_ENABLE_REG) = (1 << irql);
}


/*
 * irq disable
 */
static __inline void _irq_disable(int irql)
{
	ADM8668_INTC_REG(IRQ_DISABLE_REG) = (1<<irql);
}


/*
 * startup irq
 */
unsigned int startup_adm8668_irq(unsigned int irq)
{
	enable_adm8668_irq(irq);
	return 0;
}


/*
 * shutdown irq
 */
void shutdown_adm8668_irq(unsigned int irq)
{
	disable_adm8668_irq(irq);
}
	

/* 
 * enable 8668 irq 
 */
void enable_adm8668_irq(unsigned int irq)
{
	int s;

	/* Disable all interrupts (FIQ/IRQ) */
	s = mips_int_lock();

	if ((irq < 0) || (irq > INT_LVL_MAX)) 
		goto err_exit;

	_irq_enable(irq);

err_exit:
	/* Restore the interrupts states */ 
	mips_int_unlock(s);
}


/*
 * disable 8668 irq 
 */
void disable_adm8668_irq(unsigned int irq)
{
	int s;

	/* Disable all interrupts (FIQ/IRQ) */ 
	s = mips_int_lock();

	if ((irq < 0) || (irq > INT_LVL_MAX)) 
		goto err_exit;

	_irq_disable(irq);

err_exit:
	/* Restore the interrupts states */ 
	mips_int_unlock(s);
}


/*
 * 8668 irq ack
 */
static inline void ack_adm8668_irq(unsigned int irq_nr)
{
	ADM8668_INTC_REG(IRQ_DISABLE_REG) = (1 << irq_nr);
}


/*
 * 8668 irq end
 */
static void end_adm8668_irq(unsigned int irq_nr)
{
	ADM8668_INTC_REG(IRQ_ENABLE_REG) = (1 << irq_nr);
}


/*
 * 8668 irq set affinity 
 */
void set_affinity_adm8668_irq(unsigned int irq, unsigned long mask)
{
	return;
}

	
/*
 * system irq type
 */
static hw_irq_controller adm8668_irq_type = {
	"ADM8668 INTC",
	startup_adm8668_irq,
	shutdown_adm8668_irq,
	enable_adm8668_irq,
	disable_adm8668_irq,
	ack_adm8668_irq,
	end_adm8668_irq,
	set_affinity_adm8668_irq
};


/*
 * irq init
 */
void __init init_adm8668_irqs (void)
{
	int i;

	for (i = 0; i <= INT_LVL_MAX; i++) 
	{
		irq_desc[i].status = IRQ_DISABLED;
		irq_desc[i].action = 0;
		irq_desc[i].depth = 1;
		irq_desc[i].handler = &adm8668_irq_type;
	}
}


/*
 * system init
 */
void __init init_IRQ(void)
{
	set_except_vector(0, mipsIRQ);

	init_generic_irq();
	
	init_adm8668_irqs();
	
#ifdef CONFIG_REMOTE_DEBUG
 	
    printk("Setting debug traps - please connect the remote debugger.\n");

	
    set_debug_traps();

	
    breakpoint();

#endif


}
