/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : Daniel Liu
 *        File    : memory.c
 *        Abstract: 
 *        Date    : 2003.3.4
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/ 

#include <linux/config.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/bootmem.h>

#include <asm/bootinfo.h>
#include <asm/page.h>
#include <asm/adm8668/adm8668.h>
#include <asm/adm8668/hwprofile.h>
#include <asm/adm8668/prom.h>


struct prom_pmemblock mdesc[PROM_MAX_PMEMBLOCKS];

/* References to section boundaries */
extern char _end;
extern int brdmemsize; 

#define PFN_ALIGN(x)    (((unsigned long)(x) + (PAGE_SIZE - 1)) & PAGE_MASK)


/*
 * prom: get memory descriptor
 */
struct prom_pmemblock * __init prom_getmdesc(void)
{
    memset(mdesc, 0, sizeof(mdesc));

	mdesc[0].type = BOOT_MEM_RAM;
	mdesc[0].base = PHYSADDR(PFN_ALIGN(&_end));
#if 1	/* old code, jeansonh 060621 */
	mdesc[0].size = (64*1024*1024) - mdesc[0].base;
#else
	/* if the board memory size is less than 1MB, unreasonable
	   might not turn on CONFIG_MEMSIZE_IN_BYTES in u-boot
	   jeansonh, 060621
	 */
	if (brdmemsize < 0x00100000)
		brdmemsize = brdmemsize*(1024*1024); 

	mdesc[0].size = brdmemsize - mdesc[0].base;
#endif
	return &mdesc[0];
}


/*
 * prom: memory init
 */
void __init prom_meminit(void)
{
	struct prom_pmemblock *p;

	p = prom_getmdesc();

	while (p->size) 
	{
		add_memory_region(p->base, p->size, p->type);
        p++; 
	}
}


/*
 * prom: from prom memory
 */
void __init prom_free_prom_memory (void)
{
	int i;
	unsigned long freed = 0;
	unsigned long addr;

	for (i = 0; i < boot_mem_map.nr_map; i++) 
	{
		if (boot_mem_map.map[i].type != BOOT_MEM_ROM_DATA)
			continue;

		addr = boot_mem_map.map[i].addr;
		while (addr < boot_mem_map.map[i].addr
			      + boot_mem_map.map[i].size) 
		{
			ClearPageReserved(virt_to_page(__va(addr)));
			set_page_count(virt_to_page(__va(addr)), 1);
			free_page(__va(addr));
			addr += PAGE_SIZE;
			freed += PAGE_SIZE;
		}
	}

	printk("Freeing prom memory: %ldkb freed\n", freed >> 10);
}


