/*
 *  Copyright  2001 Flaga hf. Medical Devices, Kri Davsson <kd@flaga.is>
 *
 *  $Id: nor_map.c,v 1.1.1.1 2008/07/10 16:56:13 thkang Exp $
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 ************************************************************************/
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/jffs2.h>
#include <linux/squashfs_fs.h>
#include <asm/io.h>
//#include "nor_map.h"
#include "uboot.h"

#define	FLASH_SIZE	0x800000
#define	FLASH_PHYS_ADDR	0x10000000

static struct mtd_info *mymtd;

static __u8 adm8668_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

static __u16 adm8668_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

static __u32 adm8668_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

static void adm8668_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

static void adm8668_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

static void adm8668_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

static void adm8668_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

static void adm8668_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

static struct map_info nor_map = {
	name:		"adm8668-nor",
	size:		FLASH_SIZE,
	buswidth:	2,
	read8:		adm8668_read8,
	read16:		adm8668_read16,
	read32:		adm8668_read32,
	copy_from:	adm8668_copy_from,
	write8:		adm8668_write8,
	write16:	adm8668_write16,
	write32:	adm8668_write32,
	copy_to:	adm8668_copy_to
};

static struct mtd_partition nor_parts[] = {
	{ name: "uboot", offset: 0, size: 0x40000, mask_flags: MTD_WRITEABLE, },
	{ name: "linux", offset: 0x40000, size: 0, },
	{ name: "rootfs", offset: 0, size: 0, },
	{ name: "nvram", offset: FLASH_SIZE-(64*1024), size: (64*1024), },
	{ name: "rootfs_data", offset: 0, size: 0, },
	{ name: NULL, },
};

int __init
init_mtd_partitions(struct mtd_info *mtd, size_t size)
{
	int nr_parts = 0;
	int blocksize, off;
	size_t len;
	struct squashfs_super_block hdr;

	blocksize = mtd->erasesize;
	if (blocksize < 0x10000)
		blocksize = 0x10000;


	/* find and size linux */
	for (off = (128*1024); off < size; off += blocksize) {
		memset(&hdr, 0xe5, sizeof(hdr));

		/*
		 * Read into buffer
		 */
		if (MTD_READ(mtd, off, sizeof(hdr), &len, (char *)&hdr) ||
		    len != sizeof(hdr))
			continue;

		if ((hdr.s_magic == SQUASHFS_MAGIC) || (hdr.s_magic == SQUASHFS_MAGIC_SWAP)) {
			printk(KERN_INFO "%s: Filesystem type: squashfs, size=0x%x or be32(%x) le32(%x)\n",
				mtd->name, hdr.bytes_used,
				be32_to_cpu(hdr.bytes_used),
				le32_to_cpu(hdr.bytes_used));

			/* Update the squashfs partition size based on the superblock info */
			nor_parts[2].offset = off;
			len = hdr.bytes_used + (mtd->erasesize - 1);
			len &= ~(mtd->erasesize - 1);
			nor_parts[2].size = len;
			/* kernel ends at beginning of here. */
			nor_parts[1].size = off - nor_parts[1].offset;
			/* rootfs_data */
			nor_parts[4].offset = nor_parts[2].offset + nor_parts[2].size;
			nor_parts[4].size = nor_parts[3].offset - nor_parts[4].offset;
		}
	}
	nr_parts = 5;
	return nr_parts;
}

int __init init_nor(void)
{
	int nr_parts;

	printk(KERN_NOTICE "ADM8668 NOR flash device: %x at %x\n",
			FLASH_SIZE, FLASH_PHYS_ADDR);

	nor_map.map_priv_1 = (unsigned long)ioremap_nocache(FLASH_PHYS_ADDR,
					FLASH_SIZE);

	if (!nor_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}

	if (!(mymtd = do_map_probe("cfi_probe", &nor_map))) {
		iounmap((void *)nor_map.map_priv_1);
		return -ENXIO;
	}

	mymtd->module = THIS_MODULE;

	nr_parts = init_mtd_partitions(mymtd, mymtd->size);
	add_mtd_partitions(mymtd, nor_parts, nr_parts);

	return 0;
}

static void __exit cleanup_nor(void)
{
	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (nor_map.map_priv_1) {
		iounmap((void *)nor_map.map_priv_1);
		nor_map.map_priv_1 = 0;
	}
}

module_init(init_nor);
module_exit(cleanup_nor);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Scott Nicholas <sn@scottn.us>");
MODULE_DESCRIPTION("New MTD map driver for ADM8668 NOR Flash");
