/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : 
 *        File    : pci.c
 *        Abstract: 
 *        Date    : 
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <asm/byteorder.h>
#include <asm/pci_channel.h>
#include <asm/adm8668/adm8668.h>



volatile u32* pci_config_address_reg = (volatile u32*)KSEG1ADDR(PCICFG_BASE);
volatile u32* pci_config_data_reg = (volatile u32*)KSEG1ADDR(PCIDAT_BASE);

#define PCI_ENABLE              0x80000000
#define PCI_CMM_IOACC_EN		0x1
#define PCI_CMM_MEMACC_EN		0x2
#define PCI_CMM_MASTER_EN		0x4
#define PCI_CMM_DEF				(PCI_CMM_IOACC_EN | PCI_CMM_MEMACC_EN | PCI_CMM_MASTER_EN)

#define PCI_DEF_CACHE_LINE_SZ	0
#define PCI_DEF_LATENCY_TIMER	0x20
#define PCI_DEF_CACHE_LATENCY	((PCI_DEF_LATENCY_TIMER << 8) | PCI_DEF_CACHE_LINE_SZ)


                             
#define cfgaddr(dev, where) (((dev->bus->number & 0xff) << 0x10) |  \
                             ((dev->devfn & 0xff) << 0x08) |        \
                             (where & 0xfc)) | PCI_ENABLE


/*
 * We can't address 8 and 16 bit words directly.  Instead we have to
 * read/write a 32bit word and mask/modify the data we actually want.
 */
static int adm8668_read_config_byte (struct pci_dev *dev,
                                   int where, unsigned char *val)
{	
    *pci_config_address_reg = cfgaddr(dev, where);
    
#ifdef CONFIG_CPU_LITTLE_ENDIAN
	*val = (le32_to_cpu(*pci_config_data_reg) >> ((where&3)<<3)) & 0xff;
#else
    *val = ((*pci_config_data_reg) >> ((where&3)<<3)) & 0xff;
#endif
    
	return PCIBIOS_SUCCESSFUL;
}

static int adm8668_read_config_word (struct pci_dev *dev,
                                   int where, unsigned short *val)
{
	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
		
	*pci_config_address_reg = cfgaddr(dev, where);
	
#ifdef CONFIG_CPU_LITTLE_ENDIAN
	*val = (le32_to_cpu(*pci_config_data_reg) >> ((where&3)<<3)) & 0xffff;
#else
	*val = ((*pci_config_data_reg) >> ((where&3)<<3)) & 0xffff;
#endif

	return PCIBIOS_SUCCESSFUL;
}

int adm8668_read_config_dword (struct pci_dev *dev,
                                    int where, unsigned int *val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	*pci_config_address_reg = cfgaddr(dev, where);
	
#ifdef CONFIG_CPU_LITTLE_ENDIAN
	*val = le32_to_cpu(*pci_config_data_reg);
#else
	*val = (*pci_config_data_reg);
#endif

	return PCIBIOS_SUCCESSFUL;
}


static int adm8668_write_config_byte (struct pci_dev *dev,
                                    int where, unsigned char val)
{
	*pci_config_address_reg = cfgaddr(dev, where);
	*(volatile u8 *)(((int)pci_config_data_reg) + (where & 3)) = val;
	return PCIBIOS_SUCCESSFUL;
}


static int adm8668_write_config_word (struct pci_dev *dev,
                                    int where, unsigned short val)
{
	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	*pci_config_address_reg = cfgaddr(dev, where);
	*(volatile u16 *)(((int)pci_config_data_reg) + (where & 2)) = (val);
	return PCIBIOS_SUCCESSFUL;
}


static int adm8668_write_config_dword (struct pci_dev *dev,
                                     int where, unsigned int val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	*pci_config_address_reg = cfgaddr(dev, where);
	*pci_config_data_reg = (val);
	return PCIBIOS_SUCCESSFUL;
}


struct pci_ops adm8668_pci_ops = {
	adm8668_read_config_byte,
	adm8668_read_config_word,
	adm8668_read_config_dword,
	adm8668_write_config_byte,
	adm8668_write_config_word,
	adm8668_write_config_dword
};


struct resource pciioport_resource = {
	"PCI IO Space", 
	0x12600000,  
	0x12800000-1,
	IORESOURCE_IO
};


struct resource pciiomem_resource = {
	"PCI Memory Space", 
	0x16000000,
	0x1E000000-1,
	IORESOURCE_MEM
};


static void adm8668_pcibios_fixup(struct pci_dev *dev)
{
	printk("ADM8668 PCI FIXUP\n");
	pci_write_config_word(dev, PCI_COMMAND, PCI_CMM_DEF);
	pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, PCI_DEF_CACHE_LATENCY);
	pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, 0);
	pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, 0);

}

#ifdef CONFIG_ADM8668_DISABLE_PCI
struct pci_channel mips_pci_channels[] = {
	{ NULL, NULL, NULL , NULL , NULL}
};
#else
struct pci_channel mips_pci_channels[] = {
	{ &adm8668_pci_ops, &pciioport_resource, &pciiomem_resource,0,0xff},
	{ NULL, NULL, NULL , NULL , NULL}
};
#endif


unsigned __init int pcibios_assign_all_busses(void)
{
        return 1;
}


void __init pcibios_fixup(void)
{
	printk("PCI_BIOS_FIXUP\n");
}


void __init pcibios_fixup_irqs(void)
{
   struct pci_dev *dev;
   int slot_num;

   pci_for_each_dev(dev) {
      slot_num = PCI_SLOT(dev->devfn);
      switch(slot_num) {
         case 1: 
		 dev->irq = 14;  
		 pci_write_config_word(dev, PCI_INTERRUPT_LINE, 14);
		 break;
         case 2: 
		 dev->irq = 13;  
		 pci_write_config_word(dev, PCI_INTERRUPT_LINE, 13);
		 break;
         case 3: 
		 dev->irq = 12;  
		 pci_write_config_word(dev, PCI_INTERRUPT_LINE, 12);
		 break;
         default: break;
      }
      pci_write_config_dword(dev, PCI_BRIDGE_MASK, 0xffff0000);
   }
}


void __init pcibios_fixup_resources(struct pci_dev *dev)
{
	if (dev->devfn == 0)
	{
		adm8668_pcibios_fixup(dev);
	}
}
