/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : Daniel Liu
 *        File    : prom.c
 *        Abstract: 
 *        Date    : 2003.3.4
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/

#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/bootmem.h>
#include <asm/bootinfo.h>
#include <asm/addrspace.h>

#define USE_BUILDIN_ARGS

#define prom_envp(index) ((char *)(((int *)(int)_prom_envp)[(index)]))

extern void adm8668_serial_console_init(void);
extern void prom_meminit(void);
extern void prom_printf(char *fmt, ...);
extern void setup_prom_printf(int tty);
extern void prom_init_cmdline(void);

int prom_argc;
int *_prom_argv, *_prom_envp;
int brdmemsize = 0;

/* adm8668 platform arguments */
#define ADM8668_ARGC		5
#define ADM8668_ENVC		2

int adm8668_argc = ADM8668_ARGC;
char *adm8668_argv[ADM8668_ARGC] = {"", "rootfstype=squashfs,jffs2",
				"root=/dev/mtdblock2", "console=ttyS0", "init=/etc/preinit"};

char *adm8668_envp[ADM8668_ENVC] = {"memsize=0x04000000", NULL};

char *prom_getenv(char *envname)
{
	/*
	 * Return a pointer to the given environment variable.
	 * In 64-bit mode: we're using 64-bit pointers, but all pointers 
	 * in the PROM structures are only 32-bit, so we need some 
	 * workarounds, if we are running in 64-bit mode.
	 */
	int i, index=0;

	prom_printf("prom_getenv"); 
	if (_prom_envp == NULL)
		return (NULL);

	i = strlen(envname);

	while(prom_envp(index)) {
		if(strncmp(envname, prom_envp(index), i) == 0) {
			return(prom_envp(index+1));
		}
		index += 2;
	}

	return(NULL);
}


/*
 * initialize the prom module.
 */
int __init prom_init(int argc, char **argv, char **envp)
{
	int i, *ptr = NULL;
	/* init print from uart0 */
    	setup_prom_printf(0);

#ifdef USE_BUILDIN_ARGS
    	prom_argc = adm8668_argc;
	_prom_argv = (int *)adm8668_argv;
	_prom_envp = (int *)adm8668_envp;
#else

	if (argc > 1)
	{
		prom_argc = argc;
		_prom_argv = (int *)KSEG1ADDR(argv);
		_prom_envp = (int *)KSEG1ADDR(envp);

		/* change the address from physical addresses to kseg1 addresses 
		 * because in u-boot it uses physical addresses
		 * by jeansonh, 05252006
		 */
		ptr = &_prom_argv[1];
		for (i=1; i<argc; i++)
		{
			*ptr = KSEG1ADDR(_prom_argv[i]);
			ptr ++;
		}
	}else
	{
		prom_printf("Use default boot arguments... \n"); 
		prom_argc = adm8668_argc;
		_prom_argv = (int *)adm8668_argv;
    		_prom_envp = (int *)KSEG1ADDR(envp);
    	}
#endif
    	/* you should these macros defined in include/asm/bootinfo.h */

    	mips_machgroup = MACH_GROUP_ADM_GW;

    	mips_machtype = MACH_ADM_GW_8668;

    	/* set IO port base to zero */ 
    	set_io_port_base(0);

    	prom_printf("\nLINUX/8668 started...\n");

    	/* init command line */
    	prom_init_cmdline();

	/* read the environmental settings */
	i=0;
	while (_prom_envp[i] != (int)NULL)
	{
		if (!strncmp((char *)KSEG1ADDR(_prom_envp[i]), "memsize=", 8))
		{
			brdmemsize = simple_strtoul((char *)KSEG1ADDR(_prom_envp[i]+8), NULL, 0);
			printk("memsize=%x\n", brdmemsize);
		}
		i++;
	}

    	/* init memory map */
    	prom_meminit();

    	/* init serial console */
    	adm8668_serial_console_init();

    	return 0;
}
