/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : Jeanson Hung
 *        File    : raw_adm8668nand.h
 *        Abstract: 
 *        Date    : 
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/
#include <asm/adm8668/adm8668.h>
#ifndef __RAW_ADM8668NAND_H__
#define __RAW_ADM8668NAND_H__

#define NAND_REG_BASE	KSEG1ADDR(ADM8668_NAND_BASE)

/* define flash addresses */
#define NAND_ADDR_CONFIG	(CONFIG_ADM8668_BSP_OFFSET*1024)
#define NAND_SIZE_CONFIG	(CONFIG_ADM8668_BSP_SIZE*1024)
#define NAND_ADDR_KERNEL	0x20000 /* 384K */
#define NAND_SIZE_KERNEL	0x800000 /* 8M */
#define NAND_ADDR_JFFS2		0x820000
#define NAND_SIZE_JFFS2		0x1000000

#define NAND_SIZE_PER_BLK	0x0	// for debugging
#define NAND_PAGE_OOB_SIZE	0x10

#define CONFIG_MTD_8668_NAND_K9F5608

#ifdef CONFIG_MTD_8668_NAND_K9F5608	// 32Mx8
#define NAND_FLASH_SIZE		0x02000000	//32MB+1MB
#define NAND_PAGE_SIZE		0x200		//512+16Bytes
#define NAND_PAGE_PER_BLK	0x20		//32 pages per block
#define NAND_SIZE_PER_BLK	0x4000		//size per block; 16KB
#endif

#ifdef CONFIG_MTD_8668_NAND_K9F2808	// 16Mx8
#define NAND_FLASH_SIZE		0x01000000
#define NAND_PAGE_SIZE		0x200
#define NAND_PAGE_PER_BLK	0x20
#define NAND_SIZE_PER_BLK	0x4000
#endif

#ifdef CONFIG_MTD_8668_NAND_K9F6408	// 8Mx8
#define NAND_FLASH_SIZE		0x00800000	//8MB+512K
#define NAND_PAGE_SIZE		0x200		//512+16Bytes
#define NAND_PAGE_PER_BLK	0x10		//16 pages per block
#define NAND_SIZE_PER_BLK	0x2000
#endif

/* check NAND flash parameters */
#if NAND_SIZE_PER_BLK == 0
#error DID NOT DEFINE THE CORRECT NAND FLASH PARAMETERS!!!
#endif

#define NAND_CMD_READ0		0
#define NAND_CMD_READ1		1
#define NAND_CMD_PAGEPROG	0x10
#define NAND_CMD_READOOB	0x50
#define NAND_CMD_ERASE1		0x60
#define NAND_CMD_STATUS		0x70
#define NAND_CMD_SEQIN		0x80
#define NAND_CMD_READID		0x90
#define NAND_CMD_ERASE2		0xd0
#define NAND_CMD_RESET		0xff

#define NAND_BADBLOCK_POS	5

#define ADD1MASK	0x000000ff
#define ADD2MASK	0x0001fe00
#define ADD3MASK	0x01fe0000
#define ADD4MASK	0xfe000000
#define ADD2SHIFT	9
#define ADD3SHIFT	17
#define ADD4SHIFT	25

int nand_read(UINT8 *dst, UINT8 *src, UINT32 len);
int nand_write(UINT8 *dst, UINT8 *src, UINT32 size);
int nand_erase(UINT8 *addr, UINT32 len);

#endif

