/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : Daniel Liu
 *        File    : setup.c
 *        Abstract: 
 *        Date    : 2003.3.4
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/

#include <linux/config.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/serial.h>
#include <linux/types.h>
#include <linux/string.h>	/* for memset */
#include <asm/reboot.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/pgtable.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/adm8668/adm8668.h>

extern struct rtc_ops adm8668_rtc_ops;
extern struct rtc_ops *rtc_ops;
#ifdef CONFIG_ADM8668_USB
extern void _arc_ehci_remove(void); 
#endif
void  mips_timer_setup(struct irqaction *irq);
void  mips_time_init(void);
int adm8668_pci_module;

/* 
 * system restart
 */

void adm8668_restart(char *command)
{
  int i, s;
  
  printk(KERN_NOTICE "\n**Rebooting system...\n");
  s = mips_int_lock();
#ifdef CONFIG_ADM8668_USB
  _arc_ehci_remove();
#endif
  /* Stop Tx, Rx of LAN1 & LAN2 */
  ADM8668_LAN_REG(NETCSR6) = BIT_1|BIT_13;
  ADM8668_LAN_REG(NETCSR7) = 0;
  ADM8668_WAN_REG(NETCSR6) = BIT_1|BIT_13;
  ADM8668_WAN_REG(NETCSR7) = 0;
  
  /* reset PHY */
  ADM8668_WAN_REG(NETCSR37) = 0x20;
  for (i=0; i<10000; i++){;;}
  ADM8668_WAN_REG(NETCSR37) = 0;
  for (i=0; i<10000; i++){;;}
  
  *(unsigned int *)0xb1600000 = 0x1;    /* reset LAN1 MAC */
  *(unsigned int *)0xb1a00000 = 0x1;    /* reset LAN2 MAC */
  /**(unsigned int *)0xb1400000 = 0x0;    /* disable HW-NAT */
  *(unsigned int *)0xb1800000 = 0x1;    /* reset WLAN MAC */

  mips_int_unlock(s);
  
  /* soft reset */
  for (i=0; i<1000; i++) {;;}
  ADM8668_CONFIG_REG(ADM8668_CR1) = BIT_0;
}



/*
 * system halt
 */
void adm8668_halt(void)
{
    printk(KERN_NOTICE "\n** You can safely turn off the power\n");
    while (1);
}

/*
 * system power off
 */
void adm8668_power_off(void)
{
    adm8668_halt();
}

/* 
 * system setup
 */
void __init adm8668_setup(void)
{
	
    printk("adm8668_setup() starts.\n");
    
    board_time_init = mips_time_init;
    board_timer_setup = mips_timer_setup;

    _machine_restart = adm8668_restart;
    _machine_halt = adm8668_halt;
    _machine_power_off = adm8668_power_off;

    rtc_ops = &adm8668_rtc_ops;

    set_io_port_base(KSEG1);

    /* Disable 8668 timer */
    ADM8668_TMR_REG(TIMER0_REG_BASE, TIMER_CTRL_REG) &= ~TMR_ENABLE;
    ADM8668_INTC_REG(IRQ_DISABLE_REG) = 1 << INT_LVL_TIMERINT0;

    printk("System has PCI BIOS\n");
    adm8668_pci_module = 1;

#ifdef CONFIG_ADM8668_SPIPCM
   IFX_SPI_PCM_Diag_Init();
#endif    
}

const char *get_system_type(void)
{
	//return "ADM8668 Demo Board";
	return "Tecom - MuleAP";
}

void __init bus_error_init(void)
{
    return;
}
