/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : <ADM8668/Dignostic Program>
 *        Block   : 
 *        Creator : Sample Lin
 *        File    : spipcm.c
 *        Abstract: This file contains all constant, structure, macro, etc. defintions
 *          	    for SPI & PCM test.
 *        Date    : 10/15/2004
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *	     Sample Lin      10152004 v1.0   Created initial version 1.0
 ************************************************************************/
#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/kmod.h>
#include <linux/fs.h>
#include <linux/version.h>
#include <linux/string.h>	/* for memset */
#include <asm/io.h>
#ifdef CONFIG_MIPS_ADM8668
#include <asm/adm8668/adm8668.h>
#elif CONFIG_MIPS_AM5120
#include <asm/am5120/am5120.h>
#endif
#include "spipcm.h"



/* Private Variables */
/* IEEE802.11E test main itmes */
static char cDiagSpiPcmItem[] =  "SPI Look back | PCM Look back";

static char cDiagSpiPcmTestResult[DIAG_SPI_PCM_MAX_ITEM];

#define POST_TEST_SUCCESS       1
#define POST_TEST_FAIL          0
#define POST_TEST_UNKNOWN       0
#define TRUE                    1
#define FALSE                   0

/* Private Functions */
/* test main */
//static void DIAG_SPI_PCM_Handler(unsigned long item_num);
//static void DIAG_SPI_PCM_Clear(void);
//static void DIAG_SPI_PCM_Test_All(void);
//static char *DIAG_SPI_PCM_Test_Get(void);
//static void DIAG_SPI_PCM_Result_Post_Display(unsigned long item_num);
//static void DIAG_SPI_PCM_Enter(void);
//static void DIAG_SPI_PCM_Exit(void);

/* utility */
static void SPI_PCM_Reg_Write(unsigned long reg_type, unsigned long offset, unsigned long value);
static unsigned long SPI_PCM_Reg_Read(unsigned long reg_type, unsigned long offset);

/* test aux */
static unsigned long SPI_PCM_Reg_Toggle_TestAll(void);

static char SPI_PCM_DIAG_SPI_Lookback(void);
static char SPI_PCM_DIAG_PCM_Lookback(void);

/* ============================= Public Function =========================== */

/*-------------------------------------------------------------
 * ROUTINE NAME - IFX_SPI_PCM_Diag_Init
 *-------------------------------------------------------------
 * FUNCTION: This function will register NAT test module.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
void IFX_SPI_PCM_Diag_Init(void)
{
    
    SPI_REG_WRITE_BYTE(SPICR0, 0x06);  /* CPOL = 1, CPHA = 1 */

    //SPI_PCM_DIAG_SPI_Lookback();
    SPI_PCM_DIAG_PCM_Lookback();

} /* End of IFX_SPI_PCM_Diag_Init */




/* ============================= Handler Function =========================== */

/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Handler
 *-------------------------------------------------------------
 * FUNCTION: This function will test NAT function.
 *
 * INPUT:    item_num - test item number, base 0
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
//static void DIAG_SPI_PCM_Handler(unsigned long item_num)
//{
//    char result = POST_TEST_UNKNOWN; /* default result */


    /* test */
//    switch(item_num)
//    {
//        case SPI_PCM_TEST_SPI_LOOKBACK:
//            result = SPI_PCM_DIAG_SPI_Lookback();
//            break;

//        case SPI_PCM_TEST_PCM_LOOKBACK:
//            result = SPI_PCM_DIAG_PCM_Lookback();
//            break;

//        default:
//            printk("Err: No such test item!\n");
//            return;
//    } /* End of switch */

    /* backup test result */
//    if (result == TRUE)
//        cDiagSpiPcmTestResult[item_num] = POST_TEST_SUCCESS;
//    else if (result == FALSE)
//        cDiagSpiPcmTestResult[item_num] = POST_TEST_FAIL;
//    else
//        cDiagSpiPcmTestResult[item_num] = POST_TEST_UNKNOWN;
    /* End of if */
//} /* End of DIAG_SPI_PCM_Handler */


/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Clear
 *-------------------------------------------------------------
 * FUNCTION: This function will clear latest test result.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
static void DIAG_SPI_PCM_Clear(void)
{
    memset(cDiagSpiPcmTestResult,
           POST_TEST_UNKNOWN,
           sizeof(cDiagSpiPcmTestResult));
} /* End of DIAG_SPI_PCM_Clear */


/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Test_All
 *-------------------------------------------------------------
 * FUNCTION: This function will test all items automatically.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
//static void DIAG_SPI_PCM_Test_All(void)
//{
//    unsigned long i;


//    DIAG_SPI_PCM_Clear();

//    SPI_PCM_DIAG_SPI_Lookback();
//    SPI_PCM_DIAG_PCM_Lookback();
//} /* End of DIAG_SPI_PCM_Test_All */


/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Test_Get
 *-------------------------------------------------------------
 * FUNCTION: This function will get latest test result.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   current test result pointer
 * NOTE:
 *-------------------------------------------------------------*/
//static char *DIAG_SPI_PCM_Test_Get(void)
//{
//    return cDiagSpiPcmTestResult;
//} /* End of DIAG_SPI_PCM_Test_Get */


/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Result_Post_Display
 *-------------------------------------------------------------
 * FUNCTION: This function will display detail result description
 *           for different test item.
 *
 * INPUT:    item_num - the test result of item (base 0)
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
static void DIAG_SPI_PCM_Result_Post_Display(unsigned long item_num)
{
    switch(item_num)
    {
        default:
            break;
    } /* End of switch */
} /* End of DIAG_SPI_PCM_Result_Post_Display */


/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Enter
 *-------------------------------------------------------------
 * FUNCTION: This function will allocate resources.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
//static void DIAG_SPI_PCM_Enter(void)
//{
//} /* End of DIAG_SPI_PCM_Enter */


/*-------------------------------------------------------------
 * ROUTINE NAME - DIAG_SPI_PCM_Exit
 *-------------------------------------------------------------
 * FUNCTION: This function will free resources.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
//static void DIAG_SPI_PCM_Exit(void)
//{
//} /* End of DIAG_SPI_PCM_Exit */




/* ============================= Utility Function =========================== */

/*-------------------------------------------------------------
 * ROUTINE NAME - SPI_PCM_Reg_Write
 *-------------------------------------------------------------
 * FUNCTION: This function will write a value into a register.
 *
 * INPUT:    reg_type: NAT, LAN1, LAN2
 *           offset:   the register address offset
 *           value:    the 32-bit value
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
static void SPI_PCM_Reg_Write(unsigned long reg_type, unsigned long offset, unsigned long value)
{
    switch(reg_type)
    {
        case SPI_PCM_TYPE_SPI:
            SPI_REG_WRITE(offset, value);
            break;

        case SPI_PCM_TYPE_PCM:
            PCM_REG_WRITE(offset, value);
            break;
    } /* End of switch */
} /* End of SPI_PCM_Reg_Write */


/*-------------------------------------------------------------
 * ROUTINE NAME - SPI_PCM_Reg_Read
 *-------------------------------------------------------------
 * FUNCTION: This function will read from a register.
 *
 * INPUT:    reg_type: NAT, LAN1, LAN2
 *           offset:   the register address offset
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
static unsigned long SPI_PCM_Reg_Read(unsigned long reg_type, unsigned long offset)
{
    switch(reg_type)
    {
        case SPI_PCM_TYPE_SPI:
            return (SPI_REG_READ(offset));

        case SPI_PCM_TYPE_PCM:
            return (PCM_REG_READ(offset));
    } /* End of switch */

    return 0;
} /* End of SPI_PCM_Reg_Read */




/* ============================= Test Function =========================== */

/*-------------------------------------------------------------
 * ROUTINE NAME - SPI_PCM_Reg_Toggle_TestAll
 *-------------------------------------------------------------
 * FUNCTION: This function will test registers for NAT.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   the fail number, 0: OK
 * NOTE:
 *-------------------------------------------------------------*/
static unsigned long SPI_PCM_Reg_Toggle_TestAll(void)
{
    unsigned long val_read, val_default;
    unsigned long *reg_write_allone;
    unsigned long reg_fail = 0, mode;
    unsigned int *reg;
    unsigned int i, j, reg_num, reg_cnt;


    /* register toggle test */
    for(i=0; i<2; i++)
    {
        if (i == 0)
        {
            mode = SPI_PCM_TYPE_SPI; /* test SPI register first */
            reg = DIAG_SPI_REG;
            reg_num = DIAG_SPI_REG_NUM;
            reg_write_allone = DIAG_SPI_WRITE_ALLONE;
        }
        else
        {
            mode = SPI_PCM_TYPE_PCM; /* test PCM register later */
            reg = (unsigned int *)DIAG_PCM_REG;
            reg_num = DIAG_PCM_REG_NUM;
            reg_write_allone = DIAG_PCM_WRITE_ALLONE;
        } /* End of if */

        for(j=0; j<reg_num; j++)
        {
            if (reg[i] == 0x00)
                continue; /* dont test the register because it is read-only */
            /* End of if */

            reg_cnt = 0;

            /* get the register address & backup current register value */
            val_default = SPI_PCM_Reg_Read(mode, reg[i]);

            /* write MISC_REGISTER_TEST_PATTERN_0 = 0xffffffff */
            SPI_PCM_Reg_Write(mode, reg[i], MISC_REGISTER_TEST_PATTERN_0);
            val_read = SPI_PCM_Reg_Read(mode, reg[i]);
/*
            if ((val_read & reg_write_allone[i]) != reg_write_allone[i])
            {
                DEBUG_DIAG(DEBUG_LEVEL5,
                           ("\noffset= 0x%02x, "
                            "val_read(write all ones)= 0x%04x, "
                            "sb= 0x%04x, readback= 0x%04x",
                            reg[i], val_read, reg_write_allone[i],
                            val_default));
                reg_cnt++;
            } 
*/	   
	    /* End of if */

            /* write MISC_REGISTER_TEST_PATTERN_1 = 0x00000000 */
            SPI_PCM_Reg_Write(mode, reg[i], MISC_REGISTER_TEST_PATTERN_1);
            val_read = SPI_PCM_Reg_Read(mode, reg[i]);
/*
            if ((val_read & reg_write_allone[i]) != \
                                                MISC_REGISTER_TEST_PATTERN_1)
            {
                DEBUG_DIAG(DEBUG_LEVEL5,
                           ("\noffset= 0x%02x, "
                            "val_read(write all zero)= 0x%04x,"
                            " sb= 0x%04x, readback= 0x%04x",
                            reg[i], val_read,
                            MISC_REGISTER_TEST_PATTERN_1, val_default));
                reg_cnt++;
            }
*/
    	    /* End of if */

            /* recover the register to old value */
            SPI_PCM_Reg_Write(mode, reg[i], val_default);
            val_read = SPI_PCM_Reg_Read(mode, reg[i]);
/*
            DEBUG_DIAG(DEBUG_LEVEL5,
                       ("\noffset= 0x%02x, val_read(write readback)= 0x%04x, "
                        "readback= 0x%04x", reg[i], val_read, val_default));

            if (reg_cnt != 0)
                reg_fail++;
*/            /* End of if */
        } /* End of for */
    } /* End of for */

    return reg_fail;
} /* End of SPI_PCM_Reg_Toggle_TestAll */


/*-------------------------------------------------------------
 * ROUTINE NAME - SPI_PCM_DIAG_SPI_Lookback
 *-------------------------------------------------------------
 * FUNCTION: This function tests SPI loopback function.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:     1. Max 32B buffer.
 *-------------------------------------------------------------*/
static char SPI_PCM_DIAG_SPI_Lookback(void)
{
    char test_buf[8] = {0x12, 0x34, 0x56, 0x78, 0x9a, 0xbc, 0xde, 0xf0};
    char spicr0;
    register char i, j;


    for(i=0; i<4; i++)
    {
        /* lbm=1, ss=00, bben=0 */
        /* check {cpol, cpha} = 00, 01, 10, 11 */
        spicr0 = SPI_REG_READ_BYTE(SPICR0);
        SPI_REG_WRITE_BYTE(SPICR0, (i<<1)|0x40);
        SPI_REG_WRITE_WORD(SPITSR, 0x0ff); /* spi session start */

        for(j=0; j<8; j++)
            SPI_REG_WRITE_WORD(SPITSR, 0x300 | test_buf[j]);
        /* End of for */

        SPI_REG_WRITE_WORD(SPITSR, 0x0fe); /* spi session end */

        while(!(SPI_REG_READ_BYTE(SPISR)&0x01)); /* wait TxFIFO empty */

        for(j=0; j<8; j++)
        {
            if ((SPI_REG_READ_BYTE(SPIRDR) & 0xff) != (test_buf[j] & 0xff))
            {
                printk("\n\n\tSPI Loopback\t\t\tFAIL");
                SPI_REG_WRITE_BYTE(SPICR0, spicr0);
                return 0;
            } /* End of if */
        } /* End of for */

        SPI_REG_WRITE_BYTE(SPICR0, spicr0);
    } /* End of for */

    printk("\n\n\tSPI Loopback\t\t\tSUCCESS");
    return 1;
} /* End of SPI_PCM_DIAG_SPI_Lookback */


/*-------------------------------------------------------------
 * ROUTINE NAME - SPI_PCM_DIAG_PCM_Loopback
 *-------------------------------------------------------------
 * FUNCTION: This function tests PCM loopback function.
 *
 * INPUT:    None
 * OUTPUT:   None
 * RETURN:   None
 * NOTE:
 *-------------------------------------------------------------*/
char SPI_PCM_DIAG_PCM_Lookback(void)
{
    unsigned long chn[] = {64, 32, 16, 8, 4, 2, 1};
    unsigned long exp_rff_cnt, chk_cnt;
    register unsigned long data;
    register unsigned long i, j;


    for(i=0; i<32; i++)
    {
        PCM_REG_WRITE(PCMDR, 0xdeadbeef);
        PCM_REG_WRITE(PCMDR, 0xfeedf00d);
    } /* End of for */

    PCM_REG_WRITE_BYTE(PCMCR0, 0x01); /* invoke software reset */

        i = 7;
        
        PCM_REG_WRITE_BYTE(PCMCHN, chn[i]-1);
        PCM_REG_WRITE_BYTE(PCMDLR, 4096/(64*2*chn[i])-1);

        PCM_REG_WRITE_BYTE(PCMCR0, 0x06); /* {lbm, go, swr} = 3'b110 */
//        PCM_REG_WRITE_BYTE(PCMIIR, 0x08); /* enable RX INT */

        /* wait tx/rx done */
        if (chn[i] < 4)
            exp_rff_cnt = 1;
        else
            exp_rff_cnt = (chn[i]>>2);
        /* End of if */

        /* wait tx/rx done */
        while(PCM_REG_READ_BYTE(PCM_RFF_CNT) < exp_rff_cnt);
        PCM_REG_WRITE_BYTE(PCMCR0, 0x00); /* disable transfer */

        data = 0x04030201;
        chk_cnt = (chn[i]<4) ? 1 : (chn[i]>>2);
        for(j=0; j<chk_cnt; j++)
        {
            /* verify data */
            if (PCM_REG_READ(PCMDR) != data)
            {
                printk("\n\n\tPCM Loopback\t\t\tFAIL");
                return 0;
            } /* End of if */

            data += 0x04040404;
        } /* End of for */

        PCM_REG_WRITE_BYTE(PCMCR0, 0x01); /* invoke software reset */

 
    printk("\n\n\tPCM Loopback\t\t\tSUCCESS");
    return 1;
} /* End of SPI_PCM_DIAG_PCM_Lookback */






