/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : <ADM8668/Dignostic Program>
 *        Block   : 
 *        Creator : Sample Lin
 *        File    : spipcm.h
 *        Abstract: This file contains all constant, structure, macro, etc. defintions
 *          	    for SPI & PCM test.
 *        Date    : 10/15/2004
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *	     Sample Lin      10152004 v1.0   Created initial version 1.0
 ************************************************************************/

#include <linux/types.h>
#include <linux/string.h>	/* for memset */
#include <asm/adm8668/adm8668.h>

/* Test Definition */

#define DIAG_SPI_PCM_MAX_ITEM               2

#define SPI_PCM_TEST_SPI_LOOKBACK           0
#define SPI_PCM_TEST_PCM_LOOKBACK           1

/* Registers Toggle Test */
#define MISC_REGISTER_TEST_PATTERN_0        0xffffffff
#define MISC_REGISTER_TEST_PATTERN_1        0x00000000

#define DIAG_SPI_REG_NUM                    12
#define DIAG_PCM_REG_NUM                    9

#define SPI_PCM_TYPE_SPI                    0
#define SPI_PCM_TYPE_PCM                    1

/* SPI related CSR */
unsigned int DIAG_SPI_REG[DIAG_SPI_REG_NUM] =  
           {0x00,   /* SPICR0  Control Register 0 */
            0x04,   /* SPICR1  Control Register 1 */
            0x08,   /* SPICR2  Control Register 2 */
            0x0C,   /* SPIDLR  Divisor latch register */
            0x10,   /* SPITSR  Time-slot schedule register */
/* 0x14 */  0x00,   /* SPIRDR  Read data register */
/* 0x18 */  0x00,   /* SPISR   Status register */
            0x1C,   /* SPITTL  TxFIFO trigger level */
            0x20,   /* SPIRTL  RxFIFO trigger level */
            0x24,   /* SPIIIR  Interrupt identification register */
            0x28,   /* GPIOR   GPIO register */
            0x2C};  /* BISTR   BIST register */

static unsigned long DIAG_SPI_WRITE_ALLONE[DIAG_SPI_REG_NUM] =
   {0x000000FE,   /* SPICR0  Control Register 0 */
    0x000000FF,   /* SPICR1  Control Register 1 */
    0x000000BF,   /* SPICR2  Control Register 2 */
    0x000000FF,   /* SPIDLR  Divisor latch register */
    0x000000FF,   /* SPITSR  Time-slot schedule register */
    0x00000000,   /* SPIRDR  Read data register */
    0x00000000,   /* SPISR   Status register */
    0x0000003F,   /* SPITTL  TxFIFO trigger level */
    0x0000003F,   /* SPIRTL  RxFIFO trigger level */
    0x0000000F,   /* SPIIIR  Interrupt identification register */
    0x001FFFF8,   /* GPIOR   GPIO register */
    0x00000001};  /* BISTR   BIST register */

/* PCM related CSR */
unsigned long DIAG_PCM_REG[DIAG_PCM_REG_NUM] =  
           {0x00,   /* PCMCR0  Control Register 0 */
            0x04,   /* PCMCHN  Channel number */
            0x08,   /* PCMDLR  Divisor latch register */
            0x0C,   /* PCMTL   FIFO trigger level */
/* 0x10 */  0x00,   /* PCMSR   Status register */
            0x14,   /* PCMDR   Data register */
/* 0x18 */  0x00,   /* CHCNT   Current channel counter */
            0x1C,   /* PCMIIR  Interrupt identification register */
            0x20};  /* BISTR   BIST register */

static unsigned long DIAG_PCM_WRITE_ALLONE[DIAG_PCM_REG_NUM] =
   {0x00000003,   /* PCMCR0  Control Register 0 */
    0x0000007F,   /* PCMCHN  Channel number */
    0x000000FF,   /* PCMDLR  Divisor latch register */
    0x000000FF,   /* PCMTL   FIFO trigger level */
    0x00000000,   /* PCMSR   Status register */
    0xFFFFFFFF,   /* PCMDR   Data register */
    0x00000000,   /* CHCNT   Current channel counter */
    0x0000000C,   /* PCMIIR  Interrupt identification register */
    0x00000003};  /* BISTR   BIST register */

/* regiters read write macro */
#define SPI_REG_BASE    0x1E600000

#define SPI_REG_READ(_reg)  \
    (*((volatile unsigned long *)PA2VA(SPI_REG_BASE + _reg)))
#define SPI_REG_WRITE(_reg, _val)   \
    (*((volatile unsigned long *)PA2VA(SPI_REG_BASE + _reg)) = _val)

#define SPI_REG_READ_BYTE(_reg)  \
    (*((volatile unsigned char *)PA2VA(SPI_REG_BASE + _reg)))
#define SPI_REG_WRITE_BYTE(_reg, _val)   \
    (*((volatile unsigned char *)PA2VA(SPI_REG_BASE + _reg)) = _val)

#define SPI_REG_READ_WORD(_reg)  \
    (*((volatile unsigned short *)PA2VA(SPI_REG_BASE + _reg)))
#define SPI_REG_WRITE_WORD(_reg, _val)   \
    (*((volatile unsigned short *)PA2VA(SPI_REG_BASE + _reg)) = _val)

#define SPICR0          (0x00)
#define SPICR1          (0x04)
#define SPICR2          (0x08)
#define SPIDLR          (0x0c)
#define SPITSR          (0x10)
#define SPIRDR          (0x14)
#define SPISR           (0x18)
#define SPITTL          (0x1c)
#define SPIRTL          (0x20)
#define SPIIIR          (0x24)
#define RLYCR           (0x28)
#define BISTR           (0x2c)


#define PCM_REG_BASE    0x1E800000

#define PCM_REG_READ(_reg)  \
    (*((volatile unsigned long *)PA2VA(PCM_REG_BASE + _reg)))
#define PCM_REG_WRITE(_reg, _val)   \
    (*((volatile unsigned long *)PA2VA(PCM_REG_BASE + _reg)) = _val)

#define PCM_REG_READ_BYTE(_reg)  \
    (*((volatile unsigned char *)PA2VA(PCM_REG_BASE + _reg)))
#define PCM_REG_WRITE_BYTE(_reg, _val)   \
    (*((volatile unsigned char *)PA2VA(PCM_REG_BASE + _reg)) = _val)

#define PCM_REG_READ_WORD(_reg)  \
    (*((volatile unsigned short *)PA2VA(PCM_REG_BASE + _reg)))
#define PCM_REG_WRITE_WORD(_reg, _val)   \
    (*((volatile unsigned short *)PA2VA(PCM_REG_BASE+ _reg)) = _val)

#define PCMCR0          (0x00)
#define PCMCHN          (0x04)
#define PCMDLR          (0x08)
#define PCMTL           (0x0c)
#define PCMSR           (0x10)
#define PCMDR           (0x14)
#define CHCNT           (0x18)
#define PCMIIR          (0x1c)
#define PCMBISTR        (0x20)
/* debug register */
#define PCM_TFF_WP      (0x80)
#define PCM_TFF_RP      (0x84)
#define PCM_TFF_CNT     (0x88)
#define PCM_RFF_WP      (0x8C)
#define PCM_RFF_RP      (0x90)
#define PCM_RFF_CNT     (0x94)

/* Public Function Prototype */
void IFX_SPI_PCM_Diag_Init(void);

/* End of SpiPcm.h */
