/*
 * Carsten Langgaard, carstenl@mips.com
 * Copyright (C) 1999,2000 MIPS Technologies, Inc.  All rights reserved.
 *
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 * Setting up the clock on the MIPS boards.
 *
 */
/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 ************************************************************************/
#include <linux/types.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/hardirq.h>
#include <asm/div64.h>
#include <asm/cpu.h>
#include <asm/debug.h>
#include <asm/adm8668/adm8668.h>
#include <linux/interrupt.h>
#include <linux/timex.h>

static unsigned int r4k_offset; /* Amount to increment compare reg each time */
static unsigned int r4k_cur;    /* What counter should be at next timer irq */

extern unsigned int mips_hpt_frequency;

#define ALLINTS (IE_IRQ0 | IE_IRQ5 | STATUSB_IP9)

#define MIPS_CPU_TIMER_IRQ 7

void mips_timer_interrupt(struct pt_regs *regs)
{
	int cpu = smp_processor_id();
	int irq = MIPS_CPU_TIMER_IRQ;

#if defined(CONFIG_PREEMPT)
        preempt_disable();
#endif

	irq_enter(cpu, irq);
	/* In this handling, no need to add the irq occurency counter,
	   this is a MIPS timer exception, not a ADM8668 interrupt! 
	   Add this counter will be harmful to the real counter of LAN 
	   port (IRQ = 7)
	*/
	/* kstat.irqs[cpu][irq]++; */
	
	timer_interrupt(irq, NULL, regs);

	irq_exit(cpu, irq);

	if (softirq_pending(cpu))
		do_softirq();

#if defined(CONFIG_PREEMPT)
        for(;;)
        {
            preempt_enable_no_resched();
            if(preempt_is_disabled() || !need_resched())
                break;
                                                                                
            db_assert(intr_off());
            db_assert(!in_interrupt());
                                                                                
            preempt_disable();
            __sti();
            preempt_schedule();
            __cli();
        }
#endif

}


void __init mips_time_init(void)
{
    int adj = (ADM8668_CONFIG_REG(ADM8668_CR3) >> 11) & 0xf; 
    /* for adjustable clock selection 
       CR3 bit 14~11, 0000 -> 175Mhz, 0001 -> 180Mhz, and so on
    */
    mips_hpt_frequency = (SYS_CLOCK + adj * 5000000) /2 ;
    printk("CPU clock: %d MHz\n", (SYS_CLOCK + adj * 5000000)/1000000);
}


void __init mips_timer_setup(struct irqaction *irq)
{
    /* to generate the first timer interrupt */
	write_c0_compare(mips_hpt_frequency);
	write_c0_count(0);
	
	clear_c0_status(ST0_BEV);
	set_c0_status(ALLINTS);
}
