/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : 
 *        Block   : 
 *        Creator : Jeffrey Huang
 *        File    : 
 *        Abstract: Support thread
 *        Date    : 2005.06.03 
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *
 ************************************************************************/



#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>
#include <linux/random.h>
#include <linux/ioport.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/adm8668/adm8668.h>
#include <linux/kernel.h>
#include "tulip.h"

#include "kTulipd.h"


#ifdef TULIP_THREAD_SUPPORT

extern int tulip_thread_poll(struct net_device *dev);

struct semaphore    tulipKSemWan;
struct semaphore    tulipKSemLan;

int	wanktId;
int	lanktId;

static int tulip_daemon_wan(struct net_device *dev) 
{
    struct sched_param    sparam;

    /* not from irq context */
    if(local_irq_count(0))
        return -EPERM;

    daemonize();
    /* Block all signals except SIGTERM, SIGKILL, SIGSTOP and SIGHUP */
    spin_lock_irq(&current->sigmask_lock);
    siginitsetinv(&current->blocked, sigmask(SIGTERM)|sigmask(SIGKILL) | sigmask(SIGSTOP)| sigmask(SIGHUP));
    recalc_sigpending(current);
    spin_unlock_irq(&current->sigmask_lock);


    write_lock(&tasklist_lock);
    sprintf(current->comm,"kTulipWANd");
    write_unlock(&tasklist_lock);

    sparam.sched_priority = 10;
    sched_setscheduler(current, (int) SCHED_FIFO, &sparam);

    for(;;)
    {
        down(&tulipKSemWan); 
	tulip_thread_poll(dev);
	if (signal_pending(current) != 0)
        {
            (void)printk(KERN_NOTICE "\nTulipWan - signal received\n");
            break;
        }

    }
}


int tulip_kernel_thread_wan_init(struct net_device *dev)  
{
    sema_init(&tulipKSemWan,1);  /* init empty */

    wanktId = kernel_thread( tulip_daemon_wan ,dev, (unsigned long) CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
    if(wanktId < 0)
    {
        return wanktId;
    }
    return(wanktId);
}


static int tulip_daemon_lan(struct net_device *dev)   
{
    struct sched_param    sparam;

    /* not from irq context */
    if(local_irq_count(0))
        return -EPERM;

    daemonize();

    /* Block all signals except SIGTERM, SIGKILL, SIGSTOP and SIGHUP */
    spin_lock_irq(&current->sigmask_lock);
    siginitsetinv(&current->blocked, sigmask(SIGTERM)|sigmask(SIGKILL) | sigmask(SIGSTOP)| sigmask(SIGHUP));
    recalc_sigpending(current);
    spin_unlock_irq(&current->sigmask_lock);


    write_lock(&tasklist_lock);
    sprintf(current->comm,"kTulipLANd");
    write_unlock(&tasklist_lock);

    sparam.sched_priority = 10;
    sched_setscheduler(current, (int) SCHED_FIFO, &sparam);


    for(;;)
    {
        down(&tulipKSemLan);
        tulip_thread_poll(dev);
        if (signal_pending(current) != 0)
        {
            (void)printk(KERN_NOTICE "\nTulipLan - signal received\n");
            break;
        }

    }
}


int tulip_kernel_thread_lan_init(struct net_device *dev) 
{

    sema_init(&tulipKSemLan,1); 

    lanktId = kernel_thread( tulip_daemon_lan ,dev, (unsigned long) CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
    if(lanktId < 0)
    {
        return lanktId;
    }
    return(lanktId);
}

void kill_tulip_wan_thread()
{
        up(&tulipKSemWan);
        kill_proc(wanktId, SIGTERM, 1);
}

void kill_tulip_lan_thread()
{
        up(&tulipKSemLan);
        kill_proc(lanktId, SIGTERM, 1);
}


#endif /* TULIP_THREAD_SUPPORT */
