/************************************************************************
 *
 *    Copyright (c) 2005
 *    Infineon Technologies AG
 *    St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 *    THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED
 *    NON-EXCLUSIVE, WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE
 *    AND SUBLICENSE THIS SOFTWARE IS FREE OF CHARGE.
 *    
 *    THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY
 *    DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR
 *    IMPLIED, INCLUDING WITHOUT LIMI?TATION, WAR?RANTIES OR REPRESENTATIONS
 *    OF WORKMANSHIP, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 *    DURABILITY, THAT THE OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR
 *    FREE OR FREE OF NY THIRD PARTY CALIMS, INCLUDING WITHOUT LIMITATION
 *    CLAIMS OF THIRD PARTY INTELLECTUAL PROPERTY INFRINGEMENT.
 *    
 *    EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND
 *    EXCEPT FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE
 *    FOR ANY CLAIM OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT,
 *    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *    --------------------------------------------------------------------
 *    
 *        Project : ADM8668
 *        Block   : 
 *        Creator : Joy Lin
 *        File    : hw_nat.h
 *        Abstract: This is a include file for hw_nat.c.
 *		    It will include all constant and structure for hardware NAT
 *        Date    : 1/28/2004
 *    
 *    Modification History:
 *           By              Date     Ver.   Modification Description
 *           --------------- -------- -----  -----------------------------
 *	     Joy Lin        1/28/2004 V1.0    Initial Version
 ************************************************************************/
#ifndef __HW_NAT_H__
#define __HW_NAT_H__

#if defined(__vxworks__)
#include    "types/vxTypesOld.h"
#elif defined(__linux__)
#include <asm/adm8668/adm8668.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_nat_protocol.h>
#include <linux/netfilter_ipv4/ip_nat_core.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>

#define LAN_REG_BASE		ADM8668_LAN_BASE
#define CFG_REG_8628_BASE	ADM8668_CONFIG_BASE

#define FALSE			-1
#define TRUE			0
#endif

/******************************************************************************/
/* Bit Mask definitions                                                       */
/******************************************************************************/

#ifndef	BIT_31
    #define	BIT_31		0x80000000
#endif
#ifndef	BIT_30
    #define	BIT_30		0x40000000
#endif
#ifndef	BIT_29
    #define	BIT_29		0x20000000
#endif
#ifndef	BIT_28
    #define	BIT_28		0x10000000
#endif
#ifndef	BIT_27
    #define	BIT_27		0x08000000
#endif
#ifndef	BIT_26
    #define	BIT_26		0x04000000
#endif
#ifndef	BIT_25
    #define	BIT_25		0x02000000
#endif
#ifndef	BIT_24
    #define	BIT_24		0x01000000
#endif
#ifndef	BIT_23
    #define	BIT_23		0x00800000
#endif
#ifndef	BIT_22
    #define	BIT_22		0x00400000
#endif
#ifndef	BIT_21
    #define	BIT_21		0x00200000
#endif
#ifndef	BIT_20
    #define	BIT_20		0x00100000
#endif
#ifndef	BIT_19
    #define	BIT_19		0x00080000
#endif
#ifndef	BIT_18
    #define	BIT_18		0x00040000
#endif
#ifndef	BIT_17
    #define	BIT_17		0x00020000
#endif
#ifndef	BIT_16
    #define	BIT_16		0x00010000
#endif
#ifndef	BIT_15
    #define	BIT_15		0x00008000
#endif
#ifndef	BIT_14
    #define	BIT_14		0x00004000
#endif
#ifndef	BIT_13
    #define	BIT_13		0x00002000
#endif
#ifndef	BIT_12
    #define	BIT_12		0x00001000
#endif
#ifndef	BIT_11
    #define	BIT_11		0x00000800
#endif
#ifndef	BIT_10
    #define	BIT_10		0x00000400
#endif
#ifndef	BIT_9
    #define	BIT_9		0x00000200
#endif
#ifndef	BIT_8
    #define	BIT_8		0x00000100
#endif
#ifndef	BIT_7
    #define	BIT_7		0x00000080
#endif
#ifndef	BIT_6
    #define	BIT_6		0x00000040
#endif
#ifndef	BIT_5
    #define	BIT_5		0x00000020
#endif
#ifndef	BIT_4
    #define	BIT_4		0x00000010
#endif
#ifndef	BIT_3
    #define	BIT_3		0x00000008
#endif
#ifndef	BIT_2
    #define	BIT_2		0x00000004
#endif
#ifndef	BIT_1
    #define	BIT_1		0x00000002
#endif
#ifndef	BIT_0
    #define	BIT_0		0x00000001
#endif

// CSR offset

//--Joy+b3, joyl_01142004
#define	CSR36A		    0xd4
#define	CSR36B		    0xd8
#define	CSR36C		    0xdc
#define	CSR36D		    0xe0    // NAT queue Receive descriptor base address
#define	CSR36E		    0xe4    // NAT queue Transmit descriptor base address
#define	CSR36F		    0xe8
#define	CSR36G		    0xec
#define	CSR36H		    0xf0
#define	CSR36I		    0xf4

//--Joy+b3, joyl_01152004
#define CFG_REG(x)			        ((x) * 4)

// CR10, NAT Table Access Control
#define CR10_ERASE_BIT				BIT_2
#define CR10_WRITE_BIT				BIT_1

#define CR10_READY(x)				( x & BIT_0)
//#define CR10_WRITE(x)				((x & BIT_1) >> 1)
//#define CR10_ERASE(x)				((x & BIT_2) >> 2)
//#define CR10_INDEX(x)				((x & 0x000003F8) >> 3)

// CR17, NAT Table Access Data 6
#define CR17_ENTRY_VALID			BIT_23
#define CR17_NAT_FILTER 			BIT_22
#define CR17_FAST_PATH_ON		    BIT_21
#define CR17_FOR_TCP    			BIT_20
#define CR17_FOR_PPPOE  			BIT_19
#define CR17_FORWARD_TO_VLAN_WAN    BIT_15

// CR24, NAT Control Register
#define CR24_NATQ_EN				BIT_28
#define CR24_HW_NAT_EN				BIT_26
#define CR24_PORT1_IS_LAN			BIT_25
#define CR24_VLAN_EN				BIT_24
#define CR24_WAN_VLAN_ID			2 //see netif_cfg.c
#define CR24_LAN_VLAN_ID			1 //see netif_cfg.c

//--Joy+b3, joyl_01312004, temp
#define NAT_DESC_NUM                100

#define NATQ_BUF_SIZE               2048
#define NAT_TABLE_ENTRY_NUM         128

#define DESCRIPTOR_END_OF_RING      BIT_25

typedef struct _NAT_DESC {
        u32 	Des0;
        u32   	Des1;
        u32    	Des2;
        u32    	Des3;
        u32	nDes0;
        u32	nDes1;
        u32	nDes2;
        u32	nDes3;
} NAT_DESC, *PNAT_DESC;

#define WAIT_CR10_READY \
        while (CR10_READY(ADM8668_CONFIG_REG(ADM8668_CR10)) == 0)     

#define ERASE_WHOLE_NAT_TABLE \
        { WAIT_CR10_READY ; \
          ADM8668_CONFIG_REG(ADM8668_CR10) = CR10_ERASE_BIT; }

#define READ_NAT_TABLE_BY_INDEX(x) \
          ADM8668_CONFIG_REG(ADM8668_CR10) = ((x) << 3)

#define WRITE_NAT_TABLE_BY_INDEX(x) \
          ADM8668_CONFIG_REG(ADM8668_CR10) = (((x) << 3) | CR10_WRITE_BIT)

#define MODE_DISABLED		0
#define MODE_NORMAL		1
#define MODE_PPPOE		2
#define MODE_DISABLETURBONAT	3

#endif  // #ifndef __HW_NAT_H__
